#ifndef __OIDENTD_H_
#define __OIDENTD_H_

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#define WAIT 0x01
#define CHARSET 0x02
#define HIDEO 0x04
#define HIDEE 0x08
#define INETD 0x10
#define SPOOF 0x20
#define RANDOM 0x40
#define ALL 0x80
#define WRAPPED 0x100
#define DEBUG 0x200 
#define UID 0x400
#define GID 0x800
#define NOIDENT 0x1000
#define NUMERIC 0x2000
#define MASQ 0x4000
#define FWD 0x8000

#define PACKAGE "ojnk identd"
#define MAJOR "1"
#define MINOR "4"
#define DATE "19981017"

/* List of users allowed to spoof identd replies */
#define USERFILE "/etc/identd.spoof"

/*
** Name of file that contains the spoofed identd reply. The file should be
** in the user's home directory.
**
** Be sure to keep the leading / if you change these
*/
#define ISPOOF "/.ispoof"
#define NOIDENTF "/.noident"

#define CFILE "/proc/net/tcp"
#define MASQFILE "/proc/net/ip_masquerade"
#define MAP "/etc/oidentd.users"

/* Maximum length of identd replies */
#define MAX_ULEN 10

#define DEFAULT_PORT 113
#define DEFAULT_TIMEOUT 30
#define DEFAULT_WTIMEOUT 120
#define DEFAULT_UMASK 022

/* String prepended to a random number when the -r flag is specified */
#define UPREFIX "user"

/*
** Define this if you want random chars of the set [0-9A-Za-z] instead of
** UPREFIXxxxxx
*/
/* #define NEW_RANDOM */

#define FACILITY LOG_DAEMON
#define PRIORITY LOG_INFO
#define DPRI LOG_DEBUG

#define ERROR(x) ((flags & HIDEE) ? "UNKNOWN-ERROR" : (x))
#define OS(x) ((flags & HIDEO) ? "OTHER" : (x))
#define VALID_PORT(p) ((p != 0 && ((p & 0xffff) == p)))

int get_user(int, int, const struct in_addr *, const struct in_addr *);
int open_ispoof(const char *, const struct passwd *);
int setup_listen(int, u_long);
int service_request(int);
int fdprintf(int, const char *, ...);
void print_usage(const char *);
void print_version(void);
void main_loop(uid_t, gid_t, int, u_int, u_int, u_long);
void sigchld(int);
void sigalrm(int);
void *xmalloc(size_t);
__inline__ char drop_privs(uid_t, gid_t);
char masq(int, int, int, const struct in_addr *);
char fwd_request(int, int, int, const struct in_addr *);
char check_noident(const char *);
char check_perm(const char *);
char go_background(int);
char find_entry(u_long, char *, size_t ursz, char *, size_t);
u_char *hostlookup(u_long);
u_long get_addr(const char *);
ssize_t sockread(int, char *, size_t);
const char *fix_ident(uid_t, const char *);
int valid_number(const char *);
int read_permline(FILE *, char *, int);
#endif
